#!/usr/bin/perl
#============================================================= -*-perl-*-
#
# jsecret.pl:
#
# DESCRIPTION
#   Print out secret.pdb and secret.dat databases
#   
# AUTHOR
#   Jeff Kosowsky
#
#========================================================================
#
# COPYRIGHT 2008, 2009
# Version 0.3, released Nov 2009
# Version 0.4
#  az Thu Sep  8 15:03:06 2011
#  extended by Alexander Zangerl (az@snafu.priv.at) to handle passwords
#  between 17 and 24 characters long
# Version 0.5
#  az [2011-12-20 Tue 18:51]
#  extended by Alexander Zangerl (az@snafu.priv.at) to also handle 
#  the format used by Secret!/Android
#
#========================================================================
# 
# NOTE: This version works on both .dat and .pdb files since it matches 
# for raw records.
# In particular, it does not need the the Palm::PDB or Palm::Raw libraries
# The following header structure is assumed in order to identify each record
#	Header length is 38 bytes, consisting of:
#     4 byte start string: \xFF\@D[\cC\cb]
#     21 byte category (up to 20 characters with NULL terminator (and padding))
#     8 byte signature (IDEA encryption of first 8 bytes of garbled password
#     1 byte TAN mode flag (off/on)
#     4 byte length (32-bit long, big-endian) for size of following data
#
#========================================================================
use strict;
use warnings;

use Crypt::IDEA;
use Term::ReadKey;
use Crypt::Rijndael;
use MIME::Base64;
use Digest::MD5 qw(md5);
use File::Slurp;
use JSON;
use Getopt::Std;

use constant DATA_START  => 38;

my (%options,$source,$passwd,$cipher,$recordsref,$signature,$regmatch);
usage() if (!getopts("c:t:l",\%options)
	    or (!-f ($source=$ARGV[0]))
	    or (@ARGV>1));
my ($regtitle,$regcat); 
$regtitle=qr/$options{t}/o if ($options{t});
$regcat=qr/$options{c}/o if ($options{c});

ReadMode('noecho');
while(1) {
	printf STDERR "Please enter password: ";
	$passwd = ReadLine(0);
	chomp($passwd);
	printf STDERR "\n";
	last if $passwd ne '' || $options{l};
}
ReadMode('restore');

# first, check if this is an Android .dat file
eval { $recordsref=decode_json(read_file($source)); };
if (!$@)			# fine, it's an android JSON file
{
    # prep the comparison password
    $cipher=Crypt::Rijndael->new(md5(uc($passwd)),Crypt::Rijndael::MODE_CBC);
    $cipher->set_iv("\0" x 16 );
    $signature=encode_base64($cipher->encrypt(pad(uc($passwd)."\0")),"");

    my @readables=grep($_->{encrpasswd} eq $signature, @{$recordsref->{records}});
    die "No records match password...\n" if (!@readables);
    
    for my $r (@readables)
    {
	next if ((defined($regtitle) && $r->{title}!~$regtitle) 
		 || (defined($regcat) && $r->{category}!~$regcat));
	$regmatch=1;
	my $dec=$cipher->decrypt(decode_base64($r->{encrtext}));
	$dec=~s/\0.*//;		# remove padding
	
	print "--- title: ".$r->{title}." category: ".($r->{category}||"(none)")." length: ".length($dec)
	    ." flags:".($r->{readonly}?" readonly":"").($r->{tanflag}?" tan":"")."\n";
	print "\n$dec\n\n" if (!$options{l});
    }
    warn "No records match your regex\n" if (!$regmatch);
}
else				# not android, fall back to the old binary format then
{
    if($passwd) {
	$passwd = uc($passwd) . "\0" x (8 - length($passwd)); #Upcase & NULL pad
	$cipher = new IDEA(garble($passwd)); #Note cipher uses garbled signature
	$signature = $cipher->encrypt(substr($passwd,0,8)); #But signature doesn't
    }

    $recordsref = read_records($source, $signature);
    die "No records match password...\n" unless $#{$recordsref} >= 0;

    foreach my $record (@$recordsref) {
	my $length = get_record_info($record, \ (my $category), \ (my $tanflag));;
	next if (defined $regcat && $category!~$regcat); #No category match
	$regmatch++;
	print "--- category: $category length: $length flags: $tanflag\n";
	if (!$options{l})
	{
	    print decrypt($record, $length, $cipher) . "\n\n";
	}
    }
    warn "No record categories match regex\n" 
	if defined($regcat) && !$regmatch;
}
exit 0;

###########################################################################
sub read_records {
	my ($file, $signature) = @_;
	my $data=read_file($file);

	my @records = split(/(?=\xFF\@D[\cC\cb].{20}\0\Q$signature\E)/so, $data);
	#Note including the signature means that we only match on
	#decodable records. Undecodable records will be tacked on to the
	#end of decodable ones, but that won't hurt anything since length
	#is decoded from header
	#Note \Q..\E construct stops interpreting special chars within $signature

	shift @records; # Drop data before first record
	return \@records;
}

#Retrieve category, tanflag and length of record
sub get_record_info {
	my($rawdata, $categoryref, $tanflagref) = @_;
	$rawdata =~ /^\xFF\@D[\cC\cb]([^\0]*)\0/; 
	$$categoryref = $1;
	$$tanflagref = unpack 'c', substr($rawdata, 33, 1); # TAN mode off/on???
	return unpack("N", substr($rawdata, 34, 4));
}

#Depcrypt text of record
sub decrypt {
	my ($rawdata, $length, $cipher) = @_;
	my $rawtext=substr($rawdata,DATA_START); #Rest is the encrypted data
	my $text = "";
	for(my $i=0; $i< $length; $i+=8) {  # 8-byte blocksize
		$text .= $cipher->decrypt(substr($rawtext,$i,8));
	}
	return substr($text,0,$length);
}

# Garble password as required by older Secret! when using IDEA
# idea algorithm needs key of precisely 16 bytes, and garble expands/condenses passwords 
# between 1 and 24 bytes to this 16 byte key
#
# priming: all key slots start with zero, then password is filled in: even slots hold pass[0,1,2...]; 
#  odd slots have pass[i-3] and slot 1 stays zero. stop when out of password chars (ie. shorter than 9 chars).
# password between 9 and 16 chars: primed like above, then even slots xor'd with pass[8], pass[9]...; 
#  odd slots xor'd with slot[i-3] and slot 1 is still zero.
# password between 17 and 24 chars: primed like before, then slot 0 is xor'd with low_nibble(pass[16])
#  other even slots xor'd with (high_nibble(slot[i-1]) or low_nibble(pass[17,18...]));
#  odd slots are xor'd with high_nibble(pass[16,17,18...])
sub garble { 

	my @chars = unpack('C*',$passwd);
	my @key = (0) x 16;

	my $pwlen=@chars;
	for my $i (0..7)
	{
	    if ($i+16<$pwlen) 	# 16+ chars long
	    {
		$key[2*$i+1]= ($i==0? 0: $chars[7+$i]) ^ (0xff & ($chars[16+$i]<<4));
		$key[2*$i] = $chars[$i] ^ $chars[8+$i] ^ ($i==0? ($chars[16+$i]>>4) : (($chars[16+$i]>>4)|(($key[2*$i-1]<<4)&0xff)));
	    }
	    elsif ($i+8<$pwlen) # 8+ chars long
	    {
		$key[2*$i]=$chars[$i] ^ $chars[8+$i];
		$key[2*$i+1]=$chars[7+$i] unless ($i==0); # key[1] is fixed at 0
	    }
	    else # 1..7 chars long
	    {
		$key[2*$i]=$chars[$i];
		$key[2*$i+1]=$chars[$i-1] unless ($i==0); # key[1] is fixed at 0
	    }
	}	
	return  pack('C16',@key);
}

sub usage {
	print STDERR <<EOF;

  Usage: $0 [-l][-c regexp][-t regexp] <secret pdb or dat file>

  where:
    -l means only list item category/title/length, do not decrypt
    -t print only items with title matching the given regexp (Android only)
    -c print only items with category matching the given regexp

EOF
exit 1;
}

# pkcs5/7 padding, used by Secret!/Android
sub pad
{
    my ($input)=@_;
    my $needed=16 - (length($input) % 16);
    if ($needed!=16)
    {
	$input.=pack("c*",(($needed) x $needed));
    }
    return $input;
}
